#ifndef AABBOX_H
#define AABBOX_H

#include <QGLViewer/vec.h>

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Interface of AABBox
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class AABBox
{
public:
  inline AABBox();
  template<class T>
  AABBox(T first,const T& last);
  template<class T>
  inline AABBox& operator+=(const T& p);
  inline AABBox& operator+=(const AABBox& p);

  inline void addAABBox(const AABBox& p);
  template<class T>
  inline void addPoint(const T& p);

  inline void enlarge(const float factor);
  inline void clear();

  inline bool empty() const;
  inline const qglviewer::Vec& min() const;
  inline const qglviewer::Vec& max() const;
  inline const qglviewer::Vec& center() const;

  inline float size(unsigned short dir) const;
  inline float width() const;
  inline float length() const;
  inline float height() const;
  inline float sizeMax() const;

  inline qglviewer::Vec corner(const unsigned short i) const;
  inline qglviewer::Vec normal(const unsigned int n) const;

  void drawUnlit() const;
  void drawLit() const;

  class Face;
  inline Face face(const unsigned int pos) const;

  template <class P>
  inline bool contains(const P& p) const;
private:
  qglviewer::Vec min_;
  qglviewer::Vec max_;
  qglviewer::Vec center_;
  bool empty_;
};
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Interface of AABBox::Face
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class AABBox::Face
{
public:
  inline qglviewer::Vec corner(const unsigned int c) const;
  inline qglviewer::Vec normal() const;
  inline qglviewer::Vec center() const;
  inline qglviewer::Vec u() const;
  inline qglviewer::Vec v() const;
protected:
  friend class AABBox;
  Face(const AABBox* q,const unsigned short pos);
private:
  const AABBox*         aabbox_;
  const unsigned short  pos_;
  const unsigned short* corners_;
};
//************************************************************
// Implementation of AABBox
//************************************************************
inline
AABBox::AABBox()
  : empty_(true)
{
}
template<class T>
AABBox::AABBox(T first,const T& last)
  : empty_(true)
{
  while (first != last)
  {
    (*this) += *first++;
  }
}  
template<class T>
inline AABBox&
AABBox::operator+=(const T& p)
{
  addPoint(p);
  return *this;
}
template<class T>
inline void
AABBox::addPoint(const T& p)
{
  if (empty_)
  {
    empty_ = false;
    center_[0] = min_[0] = max_[0]  = p[0];
    center_[1] = min_[1] = max_[1]  = p[1];
    center_[2] = min_[2] = max_[2]  = p[2];
    return;
  }    
  if (p[0] < min_[0])
  {
    min_[0] = p[0];
    center_[0]=(min_[0]+max_[0])/2.0f;
  }
  else if (p[0] > max_[0])
  {
    max_[0] = p[0];
    center_[0]=(min_[0]+max_[0])/2.0f;
  }  
  if (p[1] < min_[1])
  {
    min_[1] = p[1];
    center_[1]=(min_[1]+max_[1])/2.0f;
  }
  else if (p[1] > max_[1])
  {
    max_[1] = p[1];
    center_[1]=(min_[1]+max_[1])/2.0f;
  }
  if (p[2] < min_[2])
  {
    min_[2] = p[2];
    center_[2]=(min_[2]+max_[2])/2.0f;
  }
  else if (p[2] > max_[2])
  {
    max_[2] = p[2];
    center_[2]=(min_[2]+max_[2])/2.0f;
  }
}
inline AABBox&
AABBox::operator+=(const AABBox& p)
{
  (*this) += p.min();
  (*this) += p.max();
  return *this;
}
inline void
AABBox::addAABBox(const AABBox& p)
{
  addPoint(p.min());
  addPoint(p.max());
}
void
AABBox::enlarge(const float factor)
{
  const float s0 = size(0);
  const float s1 = size(1);
  const float s2 = size(2);
  min_[0] = min_[0]-factor*s0;
  min_[1] = min_[1]-factor*s1;
  min_[2] = min_[2]-factor*s2;
  max_[0] = max_[0]+factor*s0;
  max_[1] = max_[1]+factor*s1;
  max_[2] = max_[2]+factor*s2;
}
inline void
AABBox::clear()
{
  empty_ = true;
}
inline bool
AABBox::empty() const
{
  return empty_;
}
inline const qglviewer::Vec&
AABBox::min() const
{
  return min_;
}
inline const qglviewer::Vec&
AABBox::max() const
{
  return max_;
}
inline const qglviewer::Vec&
AABBox::center() const
{
  return center_;
}
inline float
AABBox::size(unsigned short dir) const
{
  return max_[dir]-min_[dir];
}
inline float
AABBox::width() const
{
  return max_[0]-min_[0];
}
inline float
AABBox::length() const
{
  return max_[1]-min_[1];
}
inline float
AABBox::height() const
{
  return max_[2]-min_[2];
}
inline float
AABBox::sizeMax() const
{
  const float s0 = size(0);
  const float s1 = size(1);
  const float s2 = size(2);
  return s0>s1?(s0>s2?s0:s2):(s1>s2?s1:s2);
}
inline qglviewer::Vec
AABBox::corner(const unsigned short i) const
{
  return qglviewer::Vec(i%2==0    ?min_[0]:max_[0],
		(i/2)%2==0?min_[1]:max_[1],
		i/4==0    ?min_[2]:max_[2]);
}
inline qglviewer::Vec
AABBox::normal(const unsigned int n) const
{
  static const qglviewer::Vec normals[6] =
    {
	qglviewer::Vec(-1.0f, 0.0f, 0.0f),
	qglviewer::Vec(+1.0f, 0.0f, 0.0f),
	qglviewer::Vec( 0.0f,-1.0f, 0.0f),
	qglviewer::Vec( 0.0f,+1.0f, 0.0f),
	qglviewer::Vec( 0.0f, 0.0f,-1.0f),
	qglviewer::Vec( 0.0f, 0.0f,+1.0f)      
    };
  return normals[n];
}
inline AABBox::Face
AABBox::face(const unsigned int pos) const
{
  return Face(this,pos);
}
template <class P>
inline bool
AABBox::contains(const P& p) const
{
  return (p[0]>=min_[0] && p[0]<=max_[0] &&
	    p[1]>=min_[1] && p[1]<=max_[1] &&
	    p[2]>=min_[2] && p[2]<=max_[2]);
}
//************************************************************
// Implementation of AABBox::Face
//************************************************************
inline qglviewer::Vec
AABBox::Face::corner(const unsigned int c) const
{
  return aabbox_->corner(corners_[c]);
}
inline qglviewer::Vec
AABBox::Face::normal() const
{
  return aabbox_->normal(pos_);
}
inline qglviewer::Vec
AABBox::Face::center() const
{
  return (corner(0)+corner(2))/2.0f;
}
inline qglviewer::Vec
AABBox::Face::u() const
{
  return corner(1)-corner(0);
}
inline qglviewer::Vec
AABBox::Face::v() const
{
  return corner(3)-corner(0);
}

#endif // AABBOX_H
